// ***************************************************************************************
//
//	Copyright (C) 2003 Kazuhiko TAMURA. All rights reserved.
//
//	This program is free software; you can redistribute it and/or 
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//	NAME:		HashikakeRoot.java 
//	DATE:		2003.6.26	initial version
//	CREATOR:	Kazuhiko TAMURA
//
// ***************************************************************************************

package jp.gr.java_conf.ktz.puzzle.hashikake.app.view;

import java.awt.Component;
import java.awt.Frame;

import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;

import java.io.File;

import jp.gr.java_conf.ktz.puzzle.framework.ProblemInfo;

import jp.gr.java_conf.ktz.puzzle.hashikake.view.BoardContainer;
import jp.gr.java_conf.ktz.puzzle.hashikake.view.RootBoardContainer;

import jp.gr.java_conf.ktz.puzzle.hashikake.util.Command;

import jp.gr.java_conf.ktz.puzzle.hashikake.util.marshal.SaveCanceledException;

import jp.gr.java_conf.ktz.puzzle.hashikake.util.gui.GUIUtility;
import jp.gr.java_conf.ktz.puzzle.hashikake.util.gui.ComponentUsher;
import jp.gr.java_conf.ktz.puzzle.hashikake.util.gui.MenuManager;
import jp.gr.java_conf.ktz.puzzle.hashikake.util.gui.OpenProblemDialog;

import jp.gr.java_conf.ktz.puzzle.hashikake.constants.AppMenuID;

import jp.gr.java_conf.ktz.puzzle.hashikake.app.view.command.CheckResultCommand;

import jp.gr.java_conf.ktz.puzzle.hashikake.command.BoardSizeCommand;
import jp.gr.java_conf.ktz.puzzle.hashikake.command.NewCommand;
import jp.gr.java_conf.ktz.puzzle.hashikake.command.LoadSelectCommand;

/**
 *	FrameAroot viewKw
 */
public class HashikakeRoot extends RootBoardContainer {
	/** application name */
	private static final String APP_NAME = "Hash wo kakero";
	
	/**
	 *	Factory method
	 */
	public static void showRootBoard() {
		HashikakeRoot aRoot = new HashikakeRoot(new java.awt.Frame());
		aRoot.initialize();
	
		aRoot.show();
	}
	
	/**
	 *	RXgN^
	 *
	 *	@param	inContainer	wrapawt component
	 */	
	private HashikakeRoot(Frame inComponent) {
		super(inComponent);
	}
	
	/**
	 *	̃R|[lg
	 */
	protected void initializeSelf() {
		Frame aFrame = (Frame)getComponent();
		
		// ^Cg̎w
		setTitle(APP_NAME);
		
		// Menubar̍쐬
		aFrame.setMenuBar(MenuManager.createMenuManager(this));
		
		// Menuڂ𖳌ɂ
		MenuManager.currentMenuManager(this).setEnabledFor(AppMenuID.MENU_OPERATE, false);
		MenuManager.currentMenuManager(this).setEnabledFor(AppMenuID.ITEM_CLEAR, false);
		MenuManager.currentMenuManager(this).setEnabledFor(AppMenuID.ITEM_CHECK, false);
		
		// ComponentUsherɓo^
		ComponentUsher.getInstance().install(this);	
		
		// Component Tree̍\z
		java.awt.ScrollPane aScroller = new java.awt.ScrollPane(java.awt.ScrollPane.SCROLLBARS_ALWAYS);
		BoardScroller aContainer = new BoardScroller(aScroller);
		addBoard(aContainer);
	}
	
	/**
	 *	Command̎
	 *
	 *	@param	inCommand	Command
	 */
	protected void processCommandImpl(Command inCommand) {
		final Class aClass = inCommand.getClass();
		
		if (aClass == CheckResultCommand.class) {
			boolean aCorrected = ((CheckResultCommand)inCommand).isCorrected();
			
			System.out.println("checked ? : " + aCorrected);
			if (aCorrected) System.out.println("congratulations !!!");
		}
		else if (aClass == NewCommand.class) {
			HashikakeRoot.showRootBoard();		
			inCommand.consume();
		}
		else if (aClass == BoardSizeCommand.class) {
			final int aWidth = ((BoardSizeCommand)inCommand).getComponentWidth();
			final int aHeight = ((BoardSizeCommand)inCommand).getComponentHeight();
			setComponentSize(aWidth, aHeight);
			inCommand.consume();
		}
		else if (aClass == LoadSelectCommand.class) {
			OpenProblemDialog.showLoadDialog(this);	
			
			MenuManager.currentMenuManager(this).setEnabledFor(AppMenuID.MENU_OPERATE, true);
			MenuManager.currentMenuManager(this).setEnabledFor(AppMenuID.ITEM_CHECK, true);
		}
	}
	
	/**
	 *	^CgƂăZbg镶ɏC
	 *
	 *	@param inOldTitle CO̕
	 *	@return	C̕
	 */
	protected String modifiyTitleString(String inOldTitle) {
		return inOldTitle  + " - " + APP_NAME;
	}
}

